"""
Layers mapping sequences to vectors
"""

from typing import Optional

import tensorflow as tf


class YangAttention(tf.keras.layers.Layer):
    """
    Applies attention using learned variables

    Arguments
    ---------

    - `n_units` (``int``): Attention's variables units
    - `name` (``str``): Layer name


    Input shape
    -----------

    (batch_size, time_steps, channels)


    Output shape
    ------------

    (batch_size, channels)


    Examples
    --------

    .. code-block:: python3

        import tensorflow as tf
        import tavolo as tvl


        model = tf.keras.Sequential([tf.keras.layers.Embedding(vocab_size, 8, input_length=max_sequence_length),
                                     tvl.seq2vec.YangAttention()])


    References
    ----------
    `Hierarchical Attention Networks for Document Classification`_


    .. _Hierarchical Attention Networks for Document Classification: https://www.cs.cmu.edu/~./hovy/papers/16HLT-hierarchical-attention-networks.pdf

    """

    def __init__(self, n_units: Optional[int],
                 name: Optional[str] = 'yang_attention',
                 **kwargs):
        """
        Apply attention with learned weights.

        Input dimensions: (batch_size, time_steps, channels)
        Output dimensions: (batch_size, channels)

        Reference: https://www.cs.cmu.edu/~./hovy/papers/16HLT-hierarchical-attention-networks.pdf

        :param n_units: Attention's variables units
        :param name: Layer name
        """
        super(YangAttention, self).__init__(name=name, **kwargs)
        self.n_units = n_units
        self.very_small_value = (-2 ** 32 + 1)  # Used for padding to avoid attending

        # Layers
        self.omega = tf.keras.layers.Dense(self.n_units,
                                           kernel_initializer=tf.keras.initializers.RandomNormal(stddev=0.1),
                                           bias_initializer=tf.keras.initializers.RandomNormal(stddev=0.1),
                                           activation=tf.nn.tanh,
                                           dtype=self.dtype,
                                           name='omega')
        self.u_omega = self.add_variable('u_omega',
                                         shape=(self.n_units,),
                                         initializer=tf.keras.initializers.RandomNormal(stddev=0.1),
                                         dtype=self.dtype)

    def compute_mask(self, inputs, mask=None):
        return mask

    def call(self, inputs,
             mask: Optional[tf.Tensor] = None,
             **kwargs) -> tf.Tensor:
        # V
        v = self.omega(inputs)  # shape=(batch_size, time_steps, n_units)

        # VU
        vu = tf.tensordot(v, self.u_omega, axes=1, name='vu')  # shape=(batch_size, time_steps)

        # Apply masking
        if mask is not None:
            # This will make sure the padded part won't be attended
            padding = tf.ones_like(vu) * self.very_small_value  # shape=(batch_size, time_steps)
            vu = tf.where(tf.equal(mask, False), padding, vu)  # shape=(batch_size, time_steps)

        # Calculate alphas
        alphas = tf.nn.softmax(vu, name='alphas')  # shape=(batch_size, time_steps)

        # Attend
        output = tf.reduce_sum(inputs * tf.expand_dims(alphas, axis=-1), axis=1)  # shape=(batch_size, channels)

        return output

    def get_config(self):
        base_config = super().get_config()
        base_config['n_units'] = self.n_units

        return base_config

    @classmethod
    def from_config(cls, config):
        return cls(**config)



from typing import Optional

import tensorflow as tf


class MultiHeadedSelfAttention(tf.keras.layers.Layer):
    """
    Applies (multi headed) self attention, taken from the Transformer
    
    
    Arguments
    ---------
    
    - `n_heads` (``int``): Number of attention heads
    - `n_units` (``int``): Number of units (sum of units of all heads), defaults to the last dimension of the input
    - `dropout_rate` (``float``): Rate of outputs to drop in the range [0, 1]
    - `causality` (``bool``): Use causality (make each time point in output dependent only on previous timepoints of input)
    - `name` (``str``): Layer name
    
    
    Input shape
    -----------
    
    (batch_size, time_steps, channels)
    
    
    Output shape
    ------------
    
    Same shape as input.
    
    
    Examples
    --------

    Apply a 4 headed (default) self attention
    
    .. code-block:: python3
    
        import tensorflow as tf
        import tavolo as tvl
    

        model = tf.keras.Sequential([tf.keras.layers.Embedding(vocab_size, 8, input_length=max_sequence_length),
                                     tvl.seq2seq.MultiHeadedSelfAttention()])


    Apply a single headed self attention

    .. code-block:: python3

        import tensorflow as tf
        import tavolo as tvl


        model = tf.keras.Sequential([tf.keras.layers.Embedding(vocab_size, 8, input_length=max_sequence_length),
                                     tvl.seq2seq.MultiHeadedSelfAttention(n_heads=1)])

    References
    ----------
    `Attention Is All You Need`_


    .. _Attention Is All You Need:
        https://arxiv.org/abs/1706.03762
    """

    def __init__(self,
                 n_heads: Optional[int] = 4,
                 n_units: Optional[int] = None,
                 dropout_rate: Optional[float] = 0.,
                 causality: Optional[bool] = False,
                 name: Optional[str] = 'multi_headed_self_attention',
                 **kwargs):
        """
        Apply multi-headed attention

        Input dimensions: (batch_size, time_steps, channels)
        Output dimensions: (batch_size, time_steps, channels)

        Reference: https://arxiv.org/abs/1706.03762

        :param n_heads: Number of attention heads
        :param n_units: Number of units (sum of units of all heads), defaults to the last dimension of the input
        :param dropout_rate: Rate of outputs to drop in the range [0, 1]
        :param causality: Use causality (make each time point in output dependent only on previous timepoints of input)
        :param name: Layer name
        """

        super(MultiHeadedSelfAttention, self).__init__(name=name, **kwargs)

        self.n_heads = n_heads
        self.n_units = n_units
        self.dropout_rate = dropout_rate
        self.causality = causality
        self.Q = None
        self.K = None
        self.V = None
        self.dropout = tf.keras.layers.Dropout(rate=dropout_rate)
        self.very_small_value = (-2 ** 32 + 1)  # Used for padding to avoid attending

    def build(self, input_shape):
        # Units
        self.n_units = self.n_units or input_shape[-1]

        # Linear projections
        self.Q = tf.keras.layers.Dense(units=self.n_units,
                                       activation=None,
                                       name='Q',
                                       dtype=self.dtype)

        self.K = tf.keras.layers.Dense(units=self.n_units,
                                       activation=None,
                                       name='K',
                                       dtype=self.dtype)

        self.V = tf.keras.layers.Dense(units=self.n_units,
                                       activation=None,
                                       name='V',
                                       dtype=self.dtype)

    def compute_mask(self, inputs, mask=None):
        return mask

    def call(self, inputs,
             mask: Optional[tf.Tensor] = None,
             query: Optional[tf.Tensor] = None,
             query_mask: Optional[tf.Tensor] = None,
             training: Optional[bool] = False,
             **kwargs) -> tf.Tensor:
        """

        :param inputs: Inputs to be used as K and V (ex. word embeddings)
        :param mask: Mask from the previous layer
        :param query: Query to be used as Q
        :param query_mask: Mask for the query
        :param training: Is training
        """

        if query is None:
            query = inputs  # Self attention

            if query_mask is None:
                query_mask = mask

        # Linear projections
        Q = self.Q(query)  # shape=(batch_size, time_steps, n_units)
        K = self.K(inputs)  # shape=(batch_size, time_steps, n_units)
        V = self.V(inputs)  # shape=(batch_size, time_steps, n_units)

        # Split and concat
        Q = tf.concat(tf.split(Q, self.n_heads, axis=2),
                      axis=0)  # shape=(batch_size * n_heads, time_steps, n_units / n_heads)
        K = tf.concat(tf.split(K, self.n_heads, axis=2),
                      axis=0)  # shape=(batch_size * n_heads, time_steps, n_units / n_heads)
        V = tf.concat(tf.split(V, self.n_heads, axis=2),
                      axis=0)  # shape=(batch_size * n_heads, time_steps, n_units_input / n_heads)

        # Attention query
        QK = tf.matmul(Q, tf.transpose(K, perm=(0, 2, 1)))  # shape=(n_heads * batch_size, time_steps, time_steps)

        # Scale
        QK /= K.get_shape().as_list()[-1] ** 0.5  # shape=(n_heads * batch_size, time_steps, time_steps)

        # Optional key masking
        # If no mask will given a mask will be created to to represent the whole sequence minus the padding
        input_mask = mask if mask is not None else tf.sign(
            tf.abs(tf.reduce_sum(inputs, axis=-1)))  # shape=(batch_size, time_steps)
        input_mask = tf.tile(input_mask, multiples=(self.n_heads, 1))  # shape=(batch_size * n_heads, time_steps)
        input_mask = tf.tile(tf.expand_dims(input_mask, axis=1),
                             multiples=(
                                 1, tf.shape(query)[1], 1))  # shape=(batch_size * n_heads, time_steps, time_steps)
        padding = tf.ones_like(QK) * self.very_small_value  # This will make sure the padded part won't be attended
        QK = tf.where(tf.equal(input_mask, False), padding, QK)  # shape=(batch_size * n_heads, time_steps, time_steps)

        # Causality
        if self.causality:
            causality_mask = tf.ones_like(QK[0, :, :])  # shape=(time_steps, time_steps)
            causality_mask = tf.linalg.LinearOperatorLowerTriangular(
                causality_mask).to_dense()  # shape=(time_steps, time_steps)
            causality_mask = tf.tile(tf.expand_dims(  # shape=(batch_size * n_heads, time_steps, time_steps)
                causality_mask, axis=0), multiples=(tf.shape(QK)[0], 1, 1))

            padding = tf.ones_like(QK) * self.very_small_value
            QK = tf.where(tf.equal(causality_mask, False), padding,
                          QK)  # shape=(batch_size * n_heads, time_steps, time_steps)

        # Create attention weights
        alphas = tf.nn.softmax(QK)  # shape=(batch_size * n_heads, time_steps, time_steps)

        # Optional query masking
        query_mask = query_mask if query_mask is not None else tf.sign(
            tf.abs(tf.reduce_sum(query, axis=-1)))  # shape=(batch_size, time_steps)
        query_mask = tf.tile(query_mask, multiples=(self.n_heads, 1))  # shape=(batch_size * n_heads, time_steps)
        query_mask = tf.tile(tf.expand_dims(query_mask, axis=-1), multiples=(
            1, 1, tf.shape(inputs)[1]))  # shape=(batch_size * n_heads, time_steps, time_steps)
        alphas *= tf.cast(query_mask, dtype=self.dtype)  # shape=(batch_size * n_heads, time_steps, time_steps)

        # Dropout
        alphas = self.dropout(alphas, training=training)  # shape=(batch_size * n_heads, time_steps, time_steps)

        # Attend and restore shape
        outputs = tf.matmul(alphas, V)  # shape=(batch_size * n_heads, time_steps, n_units / n_heads)
        outputs = tf.concat(tf.split(outputs, self.n_heads, axis=0),
                            axis=2)  # shape=(batch_size, time_steps, n_units)

        return outputs

    def get_config(self):
        base_config = super().get_config()
        base_config['n_heads'] = self.n_heads
        base_config['n_units'] = self.n_units
        base_config['dropout_rate'] = self.dropout_rate
        base_config['causality'] = self.causality

        return base_config

    @classmethod
    def from_config(cls, config):
        return cls(**config)



from typing import Optional, Callable

import tensorflow as tf
import numpy as np


class CyclicLearningRateCallback(tf.keras.callbacks.Callback):
    """
    Apply cyclic learning rate. Supports the following scale schemes:

    - ``triangular`` - Triangular cycle
    - ``triangular2`` - Triangular cycle that shrinks amplitude by half each cycle
    - ``exp_range`` - Triangular cycle that shrinks amplitude by ``gamma ** <cycle iterations>`` each cycle


    Arguments
    ---------

    - `base_lr` (``float``): Lower boundary of each cycle
    - `max_lr` (``float``): Upper boundary of each cycle, may not be reached depending on the scaling function
    - `step_size` (``int``): Number of batches per half-cycle (step)
    - `scale_scheme` (``str``): One of ``{'triangular', 'triangular2', 'exp_range'}``. If ``scale_fn`` is passed, this argument is ignored
    - `gamma` (``float``): Constant used for the ``exp_range``'s ``scale_fn``, used as (``gamma ** <cycle iterations>``)
    - `scale_fn` (``callable``): Custom scaling policy, accepts cycle index / iterations depending on the ``scale_mode`` and must return a value in the range [0, 1]. If passed, ignores ``scale_scheme``
    - `scale_mode` (``str``): Define whether ``scale_fn`` is evaluated on cycle index or cycle iterations


    Examples
    --------

    Apply a triangular cyclic learning rate (default), with a step size of 2000 batches

    .. code-block:: python3

        import tensorflow as tf
        import tavolo as tvl

        clr = tvl.learning.CyclicLearningRateCallback(base_lr=0.001, max_lr=0.006, step_size=2000)

        model.fit(X_train, Y_train, callbacks=[clr])

    Apply a cyclic learning rate that shrinks amplitude by half each cycle

    .. code-block:: python3

        import tensorflow as tf
        import tavolo as tvl

        clr = tvl.learning.CyclicLearningRateCallback(base_lr=0.001, max_lr=0.006, step_size=2000, scale_scheme='triangular2')

        model.fit(X_train, Y_train, callbacks=[clr])

    Apply a cyclic learning rate with a custom scaling function

    .. code-block:: python3

        import tensorflow as tf
        import tavolo as tvl

        scale_fn = lambda x: 0.5 * (1 + np.sin(x * np.pi / 2))
        clr = tvl.learning.CyclicLearningRateCallback(base_lr=0.001, max_lr=0.006, step_size=2000, scale_fn=scale_fn)

        model.fit(X_train, Y_train, callbacks=[clr])


    References
    ----------

    - `Cyclical Learning Rates for Training Neural Networks`_
    - `Original implementation`_

    .. _`Cyclical Learning Rates for Training Neural Networks`: https://arxiv.org/abs/1506.01186
    .. _`Original implementation`: https://github.com/bckenstler/CLR

    """

    def __init__(self, base_lr: float = 0.001,
                 max_lr: float = 0.006,
                 step_size: int = 2000.,
                 scale_scheme: str = 'triangular',
                 gamma: float = 1.,
                 scale_fn: Optional[Callable[[int], float]] = None,
                 scale_mode: str = 'cycle'):
        """

        :param base_lr: Lower boundary of each cycle
        :param max_lr: Upper boundary of each cycle, may not be reached depending on the scaling function
        :param step_size: Number of batches per half-cycle (step)
        :param scale_scheme: One of ``{'triangular', 'triangular2', 'exp_range'}``. If ``scale_fn`` is passed,
            this argument is ignored.
        :param gamma: Constant used for the ``exp_range``'s ``scale_fn``, used as (``gamma ** <cycle iterations>``)
        :param scale_fn: Custom scaling policy, accepts cycle index / iterations depending on the ``scale_mode``
            and must return a value in the range [0, 1]. If passed, ignores ``scale_scheme``
        :param scale_mode: Define whether ``scale_fn`` is evaluated on cycle index or cycle iterations
        """

        super().__init__()

        self.base_lr = base_lr
        self.max_lr = max_lr
        self.step_size = step_size
        self.scale_scheme = scale_scheme
        self.gamma = gamma

        if not scale_fn:
            self._set_scale_scheme()
        else:
            self.scale_fn = scale_fn
            self.scale_mode = scale_mode

        # Init
        self.clr_iterations = 0
        self.trn_iterations = 0
        self.history = dict()

    def clr(self) -> float:
        """
        Calculate learning rate

        :return: Learning rate value
        """

        # Calculate learning rate
        cycle = np.floor(1 + self.clr_iterations / (2 * self.step_size))
        x = abs(self.clr_iterations / self.step_size - 2 * cycle + 1)
        clr = self.base_lr + (self.max_lr - self.base_lr) * np.maximum(0, (1 - x)) * \
              self.scale_fn(cycle if self.scale_mode == 'cycle' else self.clr_iterations)

        return clr

    def on_train_begin(self, logs: Optional[dict] = None):

        logs = logs or dict()

        # Apply learning rate to optimizer
        tf.keras.backend.set_value(self.model.optimizer.lr, self.base_lr if self.clr_iterations == 0 else self.clr())

    def on_train_batch_end(self, batch: tf.Tensor, logs: Optional[dict] = None):

        logs = logs or dict()

        # Step
        self.trn_iterations += 1
        self.clr_iterations += 1

        # Log history
        self.history.setdefault('lr', list()).append(tf.keras.backend.get_value(self.model.optimizer.lr))
        self.history.setdefault('iterations', list()).append(self.trn_iterations)

        for k, v in logs.items():
            self.history.setdefault(k, list()).append(v)

        tf.keras.backend.set_value(self.model.optimizer.lr, self.clr())

    def _set_scale_scheme(self):
        """
        Applies scale mode and function for supported (non-custom) scaling schemes

        """

        # Check for supported scale schemes
        if self.scale_scheme not in {'triangular', 'triangular2', 'exp_rage'}:
            raise ValueError('{} is not a supported scale scheme'.format(self.scale_scheme))

        # Set scheme
        if self.scale_scheme == 'triangular':
            self.scale_fn = lambda x: 1
            self.scale_mode = 'cycle'
        elif self.scale_scheme == 'triangular2':
            self.scale_fn = lambda x: 1 / (2 ** (x - 1))
            self.scale_mode = 'cycle'
        elif self.scale_scheme == 'exp_range':
            self.scale_fn = lambda x: self.gamma ** x
            self.scale_mode = 'iterations'



from typing import Optional

import tensorflow as tf


class LayerNorm(tf.keras.layers.Layer):
    """
    Apply layer normalization


    Arguments
    ---------

    - `epsilon` (``float``): Small number to avoid division by zero
    - `name` (``str``): Layer name


    Input shape
    -----------

    Arbitrary. Use the keyword argument `input_shape` (tuple of integers, does not include the samples axis) when
    using this layer as the first layer in a model.


    Output shape
    ------------

    Same shape as input.


    Examples
    --------

    .. code-block:: python3

        import tensorflow as tf
        import tavolo as tvl

        model = tf.keras.Sequential([SomeLayer(),
                                     tvl.normalization.LayerNorm()])  # Apply layer normalization on SomeLayer's output


    References
    ----------
    `Layer Normalization`_


    .. _Layer Normalization:
        https://arxiv.org/pdf/1607.06450
    """

    def __init__(self, epsilon: Optional[float] = 1e-8,
                 name: Optional[str] = 'layer_norm',
                 **kwargs):
        """
        :param epsilon: Small number to avoid division by zero
        :param name: Layer name
        """
        super(LayerNorm, self).__init__(name=name, **kwargs)

        self.epsilon = epsilon
        self.beta, self.gamma = None, None

    def build(self, input_shape):
        params_shape = input_shape[-1:]

        # Initialize beta and gamma
        self.beta = self.add_variable('beta',
                                      shape=params_shape,
                                      initializer=tf.keras.initializers.zeros,
                                      dtype=self.dtype)
        self.gamma = self.add_variable('gamma',
                                       shape=params_shape,
                                       initializer=tf.keras.initializers.ones,
                                       dtype=self.dtype)

    def compute_mask(self, inputs, mask=None):
        return mask

    def call(self, inputs,
             **kwargs) -> tf.Tensor:
        # Calculate mean and variance
        mean, variance = tf.nn.moments(inputs, axes=-1, keepdims=True)  # shape=(batch_size, 1)

        # Normalize
        normalized = (inputs - mean) / ((variance + self.epsilon) ** .5)  # shape=(batch_size, channels)

        return self.gamma * normalized + self.beta  # shape=(batch_size, channels)

    def get_config(self):
        base_config = super().get_config()
        base_config['epsilon'] = self.epsilon

        return base_config

    @classmethod
    def from_config(cls, config):
        return cls(**config)



from setuptools import setup

VERSION = '0.2.0'

setup(name='tavolo',
      version=VERSION,
      description='Collection of deep learning modules and layers for the TensorFlow framework',
      url='https://github.com/eliorc/tavolo',
      author='Elior Cohen',
      classifiers=['License :: OSI Approved :: MIT License'],
      packages=['tavolo'],
      install_requires=[
          'numpy',
          'tensorflow==2.0.0-alpha0'],
      python_requires='>=3.5')



from typing import Optional

import numpy as np
import tensorflow as tf


class PositionalEncoding(tf.keras.layers.Layer):
    """
    Create a positional encoding layer, usually added on top of an embedding layer.
    Embeds information about the position of the elements using the formula

    .. math::

        PE[pos,2i]=sin\\left(\\frac{pos}{normalize\\_factor^{\\frac{2i}{embedding\\_dim}}}\\right)

        PE[pos,2i+1]=cos\\left(\\frac{pos}{normalize\\_factor^{\\frac{2i}{embedding\\_dim}}}\\right)


    The resulting embedding gets added (point-wise) to the input.


    Arguments
    ---------

    - `max_sequence_length` (``int``): Maximum sequence length of input
    - `embedding_dim` (``int``): Dimensionality of the of the input's last dimension
    - `normalize_factor` (``float``): Normalize factor
    - `name` (``str``): Layer name


    Input shape
    -----------

    (batch_size, time_steps, channels) where time_steps equals to the ``max_sequence_length`` and channels to ``embedding_dim``


    Output shape
    ------------

    Same shape as input.


    Examples
    --------

    .. code-block:: python3

        import tensorflow as tf
        import tavolo as tvl

        model = tf.keras.Sequential([tf.keras.layers.Embedding(vocab_size, 8, input_length=max_sequence_length),
                                     tvl.embeddings.PositionalEncoding(max_sequence_length=max_sequence_length,
                                                                       embedding_dim=8)])  # Add positional encoding


    References
    ----------
    `Attention Is All You Need`_


    .. _Attention Is All You Need:
        https://arxiv.org/abs/1706.03762
    """

    def __init__(self,
                 max_sequence_length: int,
                 embedding_dim: int,
                 normalize_factor: Optional[float] = 10000,
                 name: Optional[str] = 'positional_encoding',
                 **kwargs):
        """
        :param max_sequence_length: Maximum sequence length of input
        :param embedding_dim: Dimensionality of the of the input's last dimension
        :param normalize_factor: Normalize factor
        :param name: Layer name
        """
        super().__init__(name=name, **kwargs)

        self.max_sequence_length = max_sequence_length
        self.embedding_dim = embedding_dim
        self.normalize_factor = normalize_factor

        # Error checking
        if max_sequence_length < 1:
            raise ValueError(
                'max_sequence_length must be greater than zero. (value provided {})'.format(max_sequence_length))

        if embedding_dim < 1:
            raise ValueError(
                'embedding_dim must be greater than zero. (value provided {})'.format(max_sequence_length))

        # First part of the PE function: sin and cos argument
        self.positional_encoding = np.array([
            [pos / np.power(normalize_factor, 2. * i / embedding_dim) for i in range(embedding_dim)]
            for pos in range(max_sequence_length)])

        # Second part, apply the cosine to even columns and sin to odds.
        self.positional_encoding[:, 0::2] = np.sin(self.positional_encoding[:, 0::2])
        self.positional_encoding[:, 1::2] = np.cos(self.positional_encoding[:, 1::2])

        self.positional_encoding = self.add_variable(
            'embedding_matrix',
            shape=self.positional_encoding.shape,
            initializer=tf.keras.initializers.Constant(self.positional_encoding),
            trainable=False,
            dtype=self.dtype)

    def compute_mask(self, inputs, mask=None):
        return mask

    def call(self, inputs,
             mask: Optional[tf.Tensor] = None,
             **kwargs) -> tf.Tensor:

        output = inputs + self.positional_encoding
        if mask is not None:
            output = tf.where(tf.tile(tf.expand_dims(mask, axis=-1), multiples=[1, 1, inputs.shape[-1]]), output,
                              inputs)

        return output  # shape=(batch_size, time_steps, channels)

    def get_config(self):
        base_config = super().get_config()
        base_config['max_sequence_length'] = self.max_sequence_length
        base_config['embedding_dim'] = self.embedding_dim
        base_config['normalize_factor'] = self.normalize_factor

        return base_config

    @classmethod
    def from_config(cls, config):
        return cls(**config)
